#include "types.h"
#include "riscv.h"
#include "defs.h"
#include "param.h"
#include "include/stat.h"
#include "include/spinlock.h"
#include "proc.h"
#include "fs.h"
#include "sleeplock.h"
#include "file.h"
#include "include/fcntl.h"
#include "buddy.h"


// Test 1: Simple allocation
void test_simple_allocation() {
    uint64 addr1 = buddy_alloc(1);
    if (addr1 == -1 || addr1 % PGSIZE != 0) {
        panic("Test 1 failed: Simple allocation");
    } else {
        printf("Test 1 passed: Allocated 1 page at %p\n", (void*)addr1);
    }
}

// Test 2: Allocate multiple pages
void test_multiple_pages_allocation() {
    uint64 addr2 = buddy_alloc(2);
    if (addr2 == -1 || addr2 % PGSIZE != 0) {
        panic("Test 2 failed: Multi-page allocation");
    } else {
        printf("Test 2 passed: Allocated 2 pages at %p\n", (void*)addr2);
    }
}

// Test 3: Free a page and reallocate
void test_reallocation_of_freed_page(uint64 addr1) {
    buddy_free(addr1);
    uint64 addr3 = buddy_alloc(1);
    if (addr3 != addr1) {
        panic("Test 3 failed: Reallocation of freed page");
    } else {
        printf("Test 3 passed: Freed and reallocated page at %p\n", (void*)addr3);
    }
}

// Test 4: Free and reallocate multiple pages
void test_reallocate_multiple_pages(uint64 addr2) {
    buddy_free(addr2);
    uint64 addr4 = buddy_alloc(2);
    if (addr4 != addr2) {
        panic("Test 4 failed: Reallocation of freed multi-page");
    } else {
        printf("Test 4 passed: Freed and reallocated multi-page at %p\n", (void*)addr4);
    }
}

// Test 5: Allocate a large block
void test_large_block_allocation() {
    uint64 addr5 = buddy_alloc(1 << (BUDDY_SYSTEM_SIZE - 2));
    if (addr5 == -1) {
        panic("Test 5 failed: Large block allocation");
    } else {
        printf("Test 5 passed: Allocated large block at %p\n", (void*)addr5);
    }
    buddy_free(addr5);
}

// Test 6: Allocate until exhaustion
void test_memory_exhaustion() {
    int allocations = 0;
    while (buddy_alloc(1) != -1) {
        allocations++;
    }
    printf("Test 6 passed: Exhausted memory after %d allocations\n", allocations);
}

// Intensive Test 1: Allocate and free alternately
void test_alternate_allocate_and_free() {
    printf("Starting Intensive Test 1: Alternate Allocate and Free\n");
    for (int i = 0; i < 1000; i++) {
        uint64 temp_addr = buddy_alloc(1);
        if (temp_addr == -1) {
            panic("Intensive Test 1 failed: Unexpected allocation failure");
        }
        buddy_free(temp_addr);
    }
    printf("Intensive Test 1 passed: Alternate Allocate and Free\n");
}

// Intensive Test 2: Fragmentation test
void test_fragmentation() {
    printf("Starting Intensive Test 2: Fragmentation Test\n");
    uint64 small_blocks[10];
    for (int i = 0; i < 10; i++) {
        small_blocks[i] = buddy_alloc(1);
        if (small_blocks[i] == -1) {
            panic("Intensive Test 2 failed: Unable to allocate small block");
        }
    }
    for (int i = 0; i < 5; i++) {
        buddy_free(small_blocks[i]);
    }
    uint64 large_block = buddy_alloc(4);
    if (large_block == -1) {
        panic("Intensive Test 2 failed: Unable to allocate large block after fragmentation");
    }
    printf("Intensive Test 2 passed: Allocated large block at %p\n", (void*)large_block);
    buddy_free(large_block);
    for (int i = 0; i < 10; i++) {
        if (i >= 5) buddy_free(small_blocks[i]);
    }
}

// Intensive Test 3: Random allocation and freeing
void test_random_allocation_and_freeing() {
    printf("Starting Intensive Test 3: Random Allocation and Freeing\n");
    uint64 random_addrs[100];
    for (int i = 0; i < 100; i++) {
        random_addrs[i] = buddy_alloc(1);
        if (random_addrs[i] == -1) {
            panic("Intensive Test 3 failed: Random allocation failed");
        }
    }
    for (int i = 0; i < 100; i += 2) {
        buddy_free(random_addrs[i]);
    }
    for (int i = 1; i < 100; i += 2) {
        buddy_free(random_addrs[i]);
    }
    printf("Intensive Test 3 passed: Random Allocation and Freeing\n");
}

void test_simple_allocation_with_init() {
    buddy_init(KERNBASE + (1 << 20)); // Reinitialize buddy system
    test_simple_allocation();
}

void test_multiple_pages_allocation_with_init() {
    buddy_init(KERNBASE + (1 << 20)); // Reinitialize buddy system
    test_multiple_pages_allocation();
}

void test_reallocation_of_freed_page_with_init() {
    buddy_init(KERNBASE + (1 << 20)); // Reinitialize buddy system
    uint64 addr1 = buddy_alloc(1);
    test_reallocation_of_freed_page(addr1);
}

void test_reallocate_multiple_pages_with_init() {
    buddy_init(KERNBASE + (1 << 20)); // Reinitialize buddy system
    uint64 addr2 = buddy_alloc(2);
    test_reallocate_multiple_pages(addr2);
}

void test_large_block_allocation_with_init() {
    buddy_init(KERNBASE + (1 << 20)); // Reinitialize buddy system
    test_large_block_allocation();
}

void test_memory_exhaustion_with_init() {
    buddy_init(KERNBASE + (1 << 20)); // Reinitialize buddy system
    test_memory_exhaustion();
}

void test_alternate_allocate_and_free_with_init() {
    buddy_init(KERNBASE + (1 << 20)); // Reinitialize buddy system
    test_alternate_allocate_and_free();
}

void test_fragmentation_with_init() {
    buddy_init(KERNBASE + (1 << 20)); // Reinitialize buddy system
    test_fragmentation();
}

void test_random_allocation_and_freeing_with_init() {
    buddy_init(KERNBASE + (1 << 20)); // Reinitialize buddy system
    test_random_allocation_and_freeing();
}

// Run all tests with reinitialization
void run_tests() {
    printf("Starting Buddy System Tests with Reinitialization\n");
    test_simple_allocation_with_init();
    test_multiple_pages_allocation_with_init();
    test_reallocation_of_freed_page_with_init();
    test_reallocate_multiple_pages_with_init();
    test_large_block_allocation_with_init();
    test_memory_exhaustion_with_init();
    test_alternate_allocate_and_free_with_init();
    test_fragmentation_with_init();
    test_random_allocation_and_freeing_with_init();
    printf("Buddy System Tests Finished\n");
}

